package flare.vis.controls
{
	import flare.vis.Visualization;
	import flare.vis.data.DataMode;
	import flare.vis.data.NodeMode;
	
	import flash.display.InteractiveObject;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Rectangle;
	
	/**
	 * Interactive control for dragging items. A DragControl will enable
	 * dragging of all Sprites in a container object by clicking and dragging
	 * them.
	 */
	public class DragControl extends Control
	{
		private var _cur:Sprite;
		private var _mx:Number, _my:Number;
		
		/** Indicates if drag should be followed at frame rate only.
		 *  If false, drag events can be processed faster than the frame
		 *  rate, however, this may pre-empt other processing. */
		public var trackAtFrameRate:Boolean = false;
		
		/** The active item currently being dragged. */
		public function get activeItem():Sprite { return _cur; }
		
		/**
		 * Creates a new DragControl.
		 * @param filter a Boolean-valued filter function determining which
		 *  items should be draggable.
		 */		
		public function DragControl(filter:*=null) {
			this.filter = filter;
		}
		
		/** @inheritDoc */
		public override function attach(obj:InteractiveObject):void
		{
			super.attach(obj);
			obj.addEventListener(MouseEvent.MOUSE_DOWN, onMouseDown);
		}
		
		/** @inheritDoc */
		public override function detach() : InteractiveObject
		{
			if (_object != null) {
				_object.removeEventListener(MouseEvent.MOUSE_DOWN, onMouseDown);
			}
			return super.detach();
		}
		
		private function onMouseDown(event:MouseEvent) : void {
			var s:NodeMode = event.target as NodeMode;
			
			if (s==null){
				//If not a direct NodeMode, then it might be a skin of one.
				if(event.target.hasOwnProperty('parentDocument') &&  event.target.parentDocument){
					s = event.target.parentDocument.hostComponent as NodeMode;
				}
				else if(event.target.hasOwnProperty('hostComponent')){
					s = event.target.hostComponent as NodeMode;
				}
				if(s==null){
					return; // exit if not a NodeMode
				}
			}
			
			if (_filter==null || _filter(s)) {
				_cur = s;
				_mx = _object.mouseX;
				_my = _object.mouseY;
				if (_cur is DataMode) (_cur as DataMode).fix();

				_cur.stage.addEventListener(MouseEvent.MOUSE_MOVE, onDrag);
				_cur.stage.addEventListener(MouseEvent.MOUSE_UP, onMouseUp);
				
				event.stopPropagation();
			}
		}
		
		private function onDrag(event:Event) : void {
			var x:Number = _object.mouseX;
			if (x != _mx) {
				_cur.x += (x - _mx);
				_mx = x;
			}
			
			var y:Number = _object.mouseY;
			if (y != _my) {
				_cur.y += (y - _my);
				_my = y;
			}
			
			//Dragging changes the size of the visualisation, so we have to update it.
			var vis:Visualization = _object as Visualization;			
			var r:Rectangle = new Rectangle(int(vis.bounds.x),int(vis.bounds.y),0,0);
			
			for each(var nm:NodeMode in vis.data.nodes){
				if(int(nm.x + nm.width) > r.width){
					r.width = int(nm.x+nm.width);
				}
				if(int(nm.y+nm.height) > r.height){
					r.height = int(nm.y+nm.height)
				}
			}
			vis.bounds = r;			
		}
		
		private function onMouseUp(event:MouseEvent) : void {
			if (_cur != null) {
				_cur.stage.removeEventListener(MouseEvent.MOUSE_UP, onMouseUp);
				_cur.stage.removeEventListener(MouseEvent.MOUSE_MOVE, onDrag);
				
				if (_cur is DataMode) (_cur as DataMode).unfix();
				event.stopPropagation();
			}
			_cur = null;
		}
		
	} // end of class DragControl
}