package flare.vis.controls
{
	import flare.util.Displays;
	import flare.vis.Visualization;
	
	import flash.display.InteractiveObject;
	import flash.display.Stage;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	
	/**
	 * Dispatched when the offset coordinate changes via user interaction.
	 */
	[Event(name="panned", type="flash.events.Event")]	
	
	/**
	 * Dispatched when the zoom level changes via user interaction.
	 */
	[Event(name="zoomed", type="flash.events.Event")]
	
	/**
	 * Interactive control for panning and zooming a "camera". Any sprite can
	 * be treated as a camera onto its drawing content and display list
	 * children. The PanZoomControl allows you to manipulate a sprite's
	 * transformation matrix (the <code>transform.matrix</code> property) to
	 * simulate camera movements such as panning and zooming. To pan and
	 * zoom over a collection of objects, simply add a PanZoomControl for
	 * the sprite holding the collection.
	 * 
	 * <pre>
	 * var s:Sprite; // a sprite holding a collection of items
	 * new PanZoomControl().attach(s); // attach pan and zoom controls to the sprite
	 * </pre>
	 * <p>Once a PanZoomControl has been created, panning is performed by
	 * clicking and dragging. Zooming is performed either by scrolling the
	 * mouse wheel or by clicking and dragging vertically while the control key
	 * is pressed.</p>
	 * 
	 * <p>By default, the PanZoomControl attaches itself to the
	 * <code>stage</code> to listen for mouse events. This works fine if there
	 * is only one collection of objects in the display list, but can cause
	 * trouble if you want to have multiple collections that can be separately
	 * panned and zoomed. The PanZoomControl constructor takes a second
	 * argument that specifies a "hit area", a shape in the display list that
	 * should be used to listen to the mouse events for panning and zooming.
	 * For example, this could be a background sprite behind the zoomable
	 * content, to which the "camera" sprite could be added. One can then set
	 * the <code>scrollRect</code> property to add clipping bounds to the 
	 * panning and zooming region.</p>
	 */
	public class PanZoomControl extends Control
	{
		private var px:Number, py:Number;
		private var dx:Number, dy:Number;
		private var mx:Number, my:Number;
		private var _drag:Boolean = false;
		private var _moved:Boolean = false;
		
		private var _hit:InteractiveObject;
		private var _stage:Stage;
		
		// *** CONFIG OPTIONS ***
		/**
		 * Enable panning in the horizontal. Default = true 
		 */
		public var panX:Boolean = true;
		/**
		 * Enable panning in the vertical. Default = true 
		 */
		public var panY:Boolean = true;
		/**
		 * Enable zoom. Default = true 
		 */
		public var zoomEnabled:Boolean = true;
		
		/**
		 * The zooming multiplication factor 
		 */
		public var zoomFactor:Number=0.5;
		/**
		 * Constrain the pan to only stop at the limits of the visualization
		 * (stops being able to pan the visualization off screen). Default = true 
		 */		
		public var constrainPan:Boolean = true;
		
		
		/**
		 * Maximum allowed zoom, defaults to 4x original size 
		 */
		public var maxZoom:Number=4;
		/**
		 * Minimum allowed zoom, defaults to 0.3x original size 
		 */		
		public var minZoom:Number=0.3;
		
		
		/**
		 * The amount that is left visible on a pan if it were heading off 
		 * screen and constrainPan is on
		 */
		private var panHoldMargin:Number = 80;
		
		
		/** The active hit area over which pan/zoom interactions can be performed. */
		public function get hitArea():InteractiveObject { return _hit; }
		public function set hitArea(hitArea:InteractiveObject):void {
			if (_hit != null) onRemove();
			_hit = hitArea;
			if (_object && _object.stage != null) onAdd();
		}

		[Bindable("panned")]
		/** The current offset coordinate in the visualization coordinate space. */
		public function get offset():Point
		{
			return Displays.getOffset( _object );
		}
		
		[Bindable("zoomed")]
		/** The current zoom level. */
		public function get zoomLevel():Number
		{
			return Displays.getScale( _object );
		}

		/**
		 * Creates a new PanZoomControl.
		 * @param hitArea a display object to use as the hit area for mouse
		 *  events. For example, this could be a background region over which
		 *  the panning and zooming should be done. If this argument is null,
		 *  the stage will be used.
		 */
		public function PanZoomControl( hitArea:InteractiveObject=null, onPanFunc:Function = null, onZoomFunc:Function = null):void
		{
			_hit = hitArea;
			
			if (onPanFunc  != null) this.addEventListener( "panned", onPanFunc );
			if (onZoomFunc != null) this.addEventListener( "zoomed", onZoomFunc );			
		}
		
		/** @inheritDoc */
		public override function attach(obj:InteractiveObject):void
		{
			super.attach(obj);
			if (obj != null) {
				obj.addEventListener(Event.ADDED_TO_STAGE, onAdd);
				obj.addEventListener(Event.REMOVED_FROM_STAGE, onRemove);
				if (obj.stage != null) onAdd();
			}
		}
		
		/** @inheritDoc */
		public override function detach():InteractiveObject
		{
			onRemove();
			_object.removeEventListener(Event.ADDED_TO_STAGE, onAdd);
			_object.removeEventListener(Event.REMOVED_FROM_STAGE, onRemove);
			return super.detach();
		}
		
		private function onAdd(evt:Event=null):void
		{
			_stage = _object.stage;
			if (_hit == null) {
				_hit = _stage;
			}
			_hit.addEventListener(MouseEvent.MOUSE_DOWN , onMouseDown);
			_hit.addEventListener(MouseEvent.MOUSE_WHEEL, onMouseWheel);
			_hit.addEventListener(MouseEvent.CLICK, onMouseClick, false, 50);
		}
		
		private function onRemove(evt:Event=null):void
		{
			if(_hit){
				_hit.removeEventListener(MouseEvent.MOUSE_DOWN, onMouseDown);
				_hit.removeEventListener(MouseEvent.MOUSE_WHEEL, onMouseWheel);
				_hit.removeEventListener(MouseEvent.CLICK, onMouseClick);
			}
		}
		
		private function onMouseDown(event:MouseEvent) : void
		{
			if (_stage == null) return;
			if (_hit == _object && event.target != _hit) return;

			_stage.addEventListener(MouseEvent.MOUSE_UP, onMouseUp);
			_stage.addEventListener(MouseEvent.MOUSE_MOVE, onMouseMove);
			_stage.addEventListener(MouseEvent.CLICK, onMouseClick, false, 50);
			
			px = mx = event.stageX;
			py = my = event.stageY;
			_drag = true;
		}
			
		private function onMouseMove(event:MouseEvent) : void
		{
			if (!_drag) return;
			var v:Visualization = _object as Visualization;			
			//trace('PAN VIS X='+v.x);
			var x:Number = panX?event.stageX:mx;
			var y:Number = panY?event.stageY:my;
			//Visualisation's scaled height and width
			var w:Number = v.transform.pixelBounds.width;
			var h:Number = v.transform.pixelBounds.height;		
			var zoomFactor:Number = w/v.width;
			var xAmt:Number = x-mx;
			var yAmt:Number = y-my;
		
			//trace('w='+w+' h='+h+' v.y='+v.y+' v.x='+v.x+' x='+x+' y='+y+' v.b.h='+v.bounds.height+' v.b.w='+v.bounds.width+' pbw='+v.transform.pixelBounds.width+' pbh='+v.transform.pixelBounds.height+' vis.h='+v.height+' vis.w='+v.width+' ZF='+zoomFactor+' xAmt=',xAmt,' yAmt='+yAmt)
			//trace('x='+x+' mx='+mx+' xAmt=',xAmt);
			//Hold panning from going down more than top of vis
			if(constrainPan){
				if(v.y + yAmt > v.height - panHoldMargin){
					//trace('STOP BOTTOM MOVE!');
					yAmt = v.y - (v.height - panHoldMargin);// my-v.y + (h-panHoldMargin); 
					//trace('YAMNT='+yAmt);
				}
				//Stop vis from dissapearing off the top of the area
				if(v.y + yAmt < -((v.bounds.height*zoomFactor)-panHoldMargin) ){
					//trace('STOP TOP MOVE!');
					yAmt = v.y +((v.bounds.height*zoomFactor)-panHoldMargin); 
				}
				
				//Stop panning when vis is going to dissapear off right
				if(v.x + xAmt >= (v.width - panHoldMargin)){//(v.x +(x-mx) > (w-panHoldMargin) )){
				 //	trace('STOP RIGHT MOVE!');
					xAmt = v.x -(v.width - panHoldMargin);
				}
				//Stop panning when vis is going to dissapear to the left
				if(v.x + xAmt < -(v.bounds.width*zoomFactor) + panHoldMargin){
					//trace('STOP LEFT MOVE!');
					xAmt = 0;//mx-v.x -(w-panHoldMargin); 
				}
			}
			
		
			if (!event.ctrlKey) {
				// @TODO: show the pan cursor
				
				dx = dy = NaN;
				
				//trace('PANNING v.h='+v.bounds.height+' v.w='+v.bounds.width+' v.x='+v.x+' v.y='+v.y+' my='+my+' mx='+mx+' y='+y+' y-my='+(y-my)+' PANY='+panY+' StageY='+event.stageY);
				Displays.panBy(_object, xAmt, yAmt);
				
				dispatchEvent( new Event("panned") );
			} 
			else if(zoomEnabled){
				// @TODO: show the zoom cursor
				
				if (isNaN(dx)) {
					dx = event.stageX;
					dy = event.stageY;
				}
				var dz:Number = 1 + (y-my)/100;
				Displays.zoomBy(_object, dz, dx, dy);
				
				dispatchEvent( new Event("zoomed") );
			}	
			mx = x;
			my = y;
			//trace('PAN x='+v.x+' y='+v.y);
			_moved = true;
		}
		
		private function onMouseUp(event:MouseEvent) : void
		{
			dx = dy = NaN;
			_drag = false;
			_stage.removeEventListener(MouseEvent.MOUSE_UP, onMouseUp);
			_stage.removeEventListener(MouseEvent.MOUSE_MOVE, onMouseMove);
			_stage.removeEventListener(MouseEvent.CLICK, onMouseClick);
		}
		
		private function onMouseClick(event:MouseEvent) : void
		{
			if ( _moved )
				event.stopImmediatePropagation();

			_moved = false;
		}
		
		private function onMouseWheel(event:MouseEvent) : void
		{
			if(zoomEnabled){
				var dw:Number = zoomFactor * event.delta;
				var dz:Number = dw < 0 ? 1/Math.abs(dw) : dw;
				
				var x:Number= event.stageX;
				var y:Number= event.stageY;
				var v:Visualization = _object as Visualization;
				var finalZoom:Number = v.transform.matrix.a * dz;				
				
				if ( dz != 0 && finalZoom <= maxZoom && finalZoom >= minZoom ){
					Displays.zoomBy(_object, dz,x,y);
					
					dispatchEvent( new Event("zoomed") );
				}
				//trace('ZOOM DZ='+dz+' zoomFactor='+zoomFactor+' Matrix= '+v.transform.matrix.toString());
							
			}
		}
		
	} // end of class PanZoomControl
}