package flare.vis.data.render
{
	import flare.util.Shapes;
	import flare.util.Stats;
	import flare.vis.data.DataMode;
	import flare.vis.data.EdgeMode;
	import flare.vis.data.NodeMode;
	
	import flash.display.Graphics;
	import flash.geom.Point;
	
	/**
	 * Renderer that draws edges as lines. The EdgeRenderer supports straight
	 * lines, poly lines, and curves as Bezier or cardinal splines. The type
	 * of edge drawn is determined from an EdgeMode's <code>shape</code>
	 * property and control points found in the <code>points</code> property.
	 * The line rendering properties are set by the <code>setLineStyle</code>
	 * method, which can be overridden by subclasses. By default, the line
	 * rendering settings are determined using default property values set
	 * on this class (for example, the <code>scaleMode<code> and 
	 * <code>caps</code> properties).
	 * <p>This draws a line which starts at the EdgeMode's X/Y, so if you do a 'lineTo()' 
	 * call within here you'll need to give it co-ordinates relative to e.x/e.y or e.x1/e.y1</p> 
	 */
	public class EdgeRenderer implements IRenderer
	{		
		private static const ROOT3:Number = Math.sqrt(3);
		
		private static var _instance:EdgeRenderer = new EdgeRenderer();
		/** Static EdgeRenderer instance. */
		public static function get instance():EdgeRenderer { return _instance; }
		
		/** Pixel hinting flag for line rendering. */
		public var pixelHinting:Boolean = false;
		/** Scale mode for line rendering (normal by default). */
		public var scaleMode:String = "normal";
		/** The joint style for line rendering. */
		public var joints:String = null;
		/** The caps style for line rendering. */
		public var caps:String = null;
		/** The miter limit for line rendering. */
		public var miterLimit:int = 3;	
		// temporary variables			
		private var _q:Point = new Point();
		private var _pts:Array = new Array(20);
		
		//Deltas
		private var dx:Number;
		private var dy:Number;
		private var dd:Number;
		
		//Start and end temps
		private var x1:Number;
		private var	y1:Number;			
		private var x2:Number; 
		private var y2:Number; 
		
		//Start and end Points
		private var startPoint:Point;
		private var endPoint:Point;
		
		//Graphics
		private var g:Graphics;
		
		//Controls
		private var ctrls:Array;
		
		//The Edge Mode
		private var e:EdgeMode;
		
		/** @inheritDoc */
		public function render(d:DataMode):void
		{
			e = d as EdgeMode;
			if (e == null) { return; } 
				
			//trace('EDGE SOURCE H='+s.height+' W='+s.width+' TARGET H='+t.height+' W='+t.width);			
			g = e.graphics;			
			ctrls = e.points as Array;
			
			//Start and end points
			x1 = e.x1; 
			y1 = e.y1;			
			x2 = e.x2; 
			y2 = e.y2;
						
			//Set the EdgeMode's position to be at the start of the edge. 
			//This makes everything following relative to a 0,0 start position.		
			d.x = x1;
			d.y = y1;
			
			//Set start and end points
			startPoint = new Point();
			startPoint.x = 0;
			startPoint.y = 0;			
			endPoint = new Point();
			endPoint.x = -(x2-x1);
			endPoint.y = y2-y1;
			
			//Adjust the end points to not run into Node graphics
			//(If the option has been set to do so)
			adjustEndPoints(e.source,e.target);

			//Draw the line
			drawLine();

			//Get unit vector along arrow line 
			//Recalculating here after the line's start and end points have been adjusted. 
			dx = startPoint.x - endPoint.x;// - startPoint.x; 
			dy = startPoint.y + endPoint.y;// - startPoint.y;
			dd = Math.sqrt(dx*dx + dy*dy);
			dx /= dd; 
			dy /= dd;
			
			//trace('DRAWING LINE FROM ['+startPoint.x+','+startPoint.y+'] TO ['+endPoint.x+','+endPoint.y+']');
			//If we have a starting line shape, draw that
			drawStartShape();
			//If we have an endPoint, draw that
			drawEndShape();			
		}
		
		
		/**
		 *  Now we're positioning the end of the line to sit in a 'nice' position around the node's 
		 *	visual representation.
		 *	Also, will optionally move the start of the line in the same manner, although this is configurable
		 *	as this can cause issues with some layouts.
		 *	
		 *	First, work out which direction the arrow is coming from by looking at the dx/dy vars
		 *	 +ve dy = down, -ve dy = up
		 *	 +ve dx = right, -ve dx = left
		 *	 higher the number (closer to 1), the steeper the angle. 
		 * @param s
		 * @param t
		 * 
		 */
		protected function adjustEndPoints(source:NodeMode,target:NodeMode):void{				
			//First, let's apply any simple offsets we've been given.
			startPoint.x += e.startXOffset;
			startPoint.y += e.startYOffset;
			endPoint.x += e.endXOffset;
			endPoint.y += e.endYOffset;
			
			if(source.dodgeIconOutgoing || target.dodgeIconIncoming){										
				
				// get unit vector along arrow line 
				dx = -endPoint.x;// - startPoint.x; 
				dy = endPoint.y;// - startPoint.y;
				dd = Math.sqrt(dx*dx + dy*dy);
				dx /= dd; 
				dy /= dd;
				if(target.height>0){
					//Line is coming down from source to target
					if(dy > 0.5){
						if(target.dodgeIconIncoming){						 						
							endPoint.y -= target.height/2;
						}
						if(source.dodgeIconOutgoing){ 
							//trace('Moving line down');
							startPoint.y += source.height/1.2;
						}
					}
					else if(dy < -0.5){//Line is coming up from source						
						if(source.dodgeIconOutgoing){ 
							//trace('Moving line up');
							startPoint.y -= source.height/1.2;
						}
						if(target.dodgeIconIncoming){ 
							endPoint.y += source.height/2;
						}
					}					
				}
				
				if(target.width>0){
					//Arrow running left to right
					if(dx > 0.5){
						if(target.dodgeIconIncoming){							
							endPoint.x += target.width/2;
						}
						if(source.dodgeIconOutgoing){ 
							//trace('Moving line left');
							startPoint.x -= source.width/1.2;
						}
					}						
					else if(dx < -0.5){//right to left
						if(target.dodgeIconIncoming){ 						
							endPoint.x -= target.width/2;
						}
						if(source.dodgeIconOutgoing){ 
							//trace('Moving line right');
							startPoint.x += source.width/1.2;
						}
					}
				}	
			}
		}
		
		
		protected function drawStartShape():void{
			//Do we have a line start Shape?
			if(e.startShapeType == LineShapeType.NONE){
				return;
			}
						
			// determine arrow head size
			var ah:Number = e.startShapeHeight;
			var aw:Number = e.startShapeWidth/2;
			
			if (ah < 0 && aw < 0)	aw = 1.5 * e.lineWidth;
			if (ah < 0) 			ah = ROOT3 * aw;
			else if (aw < 0) 		aw = ah / ROOT3;
			
			//trace('   ARROW TIP at ',_p+' Parent ['+e.source.data.name+'] Source ['+e.target.data.name+'] height='+ah+' width='+aw);									
			// set final point positions
			dd = e.lineWidth/2;					
		
			
			
			if(e.startShapeType == LineShapeType.FILLEDCIRCLE){
				Shapes.drawCircle(g,e.startShapeWidth);
			}

			// IF ARROW, DRAW!
			if (e.startShapeType == LineShapeType.FILLEDARROW || e.startShapeType == LineShapeType.OPENARROW ) {					
				//trace('DRAWING START ARROW!!!');
				//Arrow points
				x1 = (startPoint.x + ah*dx) - aw*dy; 
				y1 = (startPoint.y - ah*dy) - aw*dx;
				
				x2 = (startPoint.x + ah*dx) + aw*dy; 
				y2 = (startPoint.y - ah*dy) + aw*dx;												
				
				//trace('DRAWING ARROW ['+x1+','+y1+'] ['+startPoint.x+','+startPoint.y+'] : ['+x2+','+y2+'] FROM dx='+dx+' dy='+dy);
				if (e.startShapeType == LineShapeType.FILLEDARROW) {
					g.lineStyle();					
					g.moveTo(x1, y1);
					g.beginFill(e.lineColor, e.lineAlpha);
					g.lineTo(startPoint.x, startPoint.y);
					g.lineTo(x2, y2);
					g.endFill();
				} else if (e.startShapeType == LineShapeType.OPENARROW) {
					// offset the anchor point (the end point for the edge connector)
					// so that edge doesn't "overshoot" the arrow head
					startPoint.x -= dd*dx;
					startPoint.y -= dd*dy;
					dd += e.lineWidth;
					dd = ah - dd;		
					
					g.moveTo(x1, y1);
					g.lineTo(startPoint.x, startPoint.y);
					g.lineTo(x2, y2);
				}
			}
		}
		
		protected function drawEndShape():void{
			//Do we have a line start Shape?
			if(e.endShapeType == LineShapeType.NONE){
				return;
			}

			if(e.endShapeType == LineShapeType.FILLEDCIRCLE || e.endShapeType == LineShapeType.OPENCIRCLE){
				
				//Arrow points
				x1 = endPoint.x + e.endShapeWidth*dx;
				y1 = endPoint.y - e.endShapeWidth*dy;
				if(e.endShapeType == LineShapeType.FILLEDCIRCLE){ 
					g.beginFill(e.lineColor, e.lineAlpha);
					g.drawCircle(x1, y1, e.endShapeWidth);
					g.endFill();
				}
				else{
					g.drawCircle(x1, y1, e.endShapeWidth);
				}
			}
			
			// IF ARROW, DRAW!
			if (e.endShapeType == LineShapeType.FILLEDARROW || e.endShapeType == LineShapeType.OPENARROW ) {					
				
				// determine arrow head size
				var ah:Number = e.endShapeHeight;
				var aw:Number = e.endShapeWidth/2;
				
				if (ah < 0 && aw < 0)	aw = 1.5 * e.lineWidth;
				if (ah < 0) 			ah = ROOT3 * aw;
				else if (aw < 0) 		aw = ah / ROOT3;
				
				//trace('   ARROW TIP at ',_p+' Parent ['+e.source.data.name+'] Source ['+e.target.data.name+'] height='+ah+' width='+aw);									
				// set final point positions
				dd = e.lineWidth/2;		
				
				
				//Arrow points
				x1 = (endPoint.x + ah*dx) - aw*dy; 
				y1 = (endPoint.y - ah*dy) - aw*dx;
				
				x2 = (endPoint.x + ah*dx) + aw*dy; 
				y2 = (endPoint.y - ah*dy) + aw*dx;												
				
				//	trace('DRAWING ARROW ['+x1+','+y1+'] ['+endPoint.x+','+endPoint.y+'] : ['+x2+','+y2+'] FROM dx='+dx+' dy='+dy);
				if (e.endShapeType == LineShapeType.FILLEDARROW) {
					g.lineStyle();					
					g.moveTo(x1, y1);
					g.beginFill(e.lineColor, e.lineAlpha);
					g.lineTo(endPoint.x, endPoint.y);
					g.lineTo(x2, y2);
					g.endFill();
				} else if (e.endShapeType == LineShapeType.OPENARROW) {
					// offset the anchor point (the end point for the edge connector)
					// so that edge doesn't "overshoot" the arrow head
					endPoint.x -= dd*dx;
					endPoint.y -= dd*dy;
					dd += e.lineWidth;
					//endPoint.x += t.width/2;				
					dd = ah - dd;		
					
					
					g.moveTo(x1, y1);
					g.lineTo(endPoint.x, endPoint.y);
					g.lineTo(x2, y2);
				}
			}
		}
		
		
		
		/**
		 * Sets the line style for edge rendering.
		 * @param e the EdgeMode to render
		 * @param g the Graphics context to draw with
		 */
		protected function setLineStyle():void
		{
			var lineAlpha:Number = e.lineAlpha;
			if (lineAlpha == 0) return;
			
			g.lineStyle(e.lineWidth, e.lineColor, lineAlpha, 
				pixelHinting, scaleMode, caps, joints, miterLimit);
		}
		
		
		protected function drawLine():void{
			// draw the edge
			g.clear(); // clear it out
			if(e.source.dodgeIconOutgoing){
				g.moveTo(startPoint.x,startPoint.y);
			}
			setLineStyle(); // set the line style						
			//If we haven't been given control points, create default mid point	
			//Only used if a shaped (curve) line (only used for shaped line below)			
			if((e.shape == Shapes.BEZIER || e.shape == Shapes.CARDINAL ||e.shape == Shapes.BSPLINE) && !ctrls){				
				var cX:Number = startPoint.x  + ((endPoint.x-startPoint.x )/2);
				var cY:Number;
				var yOffset:Number = dx > 0.5?40:-40;				
				cY = startPoint.y + ((endPoint.y-startPoint.y)/2) + yOffset;												
			}
			
			if (e.shape == Shapes.BEZIER){// && ctrls != null && ctrls.length > 1) {								
				ctrls = ctrls?ctrls:[40,40];			
				if (ctrls.length < 4)
				{
					//Shapes.drawCurve(g, startPoint.x, startPoint.y, ctrls[0],ctrls[1], endPoint.x, endPoint.y);
					trace('CURVE FROM X='+startPoint.x+' Y='+startPoint.y+' TO X='+endPoint.x+' Y='+endPoint.y+' THROUGH x='+ctrls[0]+' y='+ctrls[1]);
					//g.moveTo(startPoint.x,startPoint.y);
					g.curveTo(cX,cY,endPoint.x, endPoint.y);
				}
				else
				{
					Shapes.drawCubic(g, x1, y1, ctrls[0], ctrls[1],	ctrls[2], ctrls[3], x2, y2);
				}
			}
			else if (e.shape == Shapes.CARDINAL){
				ctrls = ctrls?ctrls:[cX,cY];
				Shapes.consolidate(x1, y1, ctrls, x2, y2, _pts);
				Shapes.drawCardinal(g, _pts, 2+ctrls.length/2);
			}
			else if (e.shape == Shapes.BSPLINE){
				ctrls = ctrls?ctrls:[cX,cY];
				Shapes.consolidate(startPoint.x, startPoint.y, ctrls, endPoint.x, endPoint.y, _pts);
				Shapes.drawBSpline(g, _pts, 2+ctrls.length/2,false);
			}			
			else{ //ELSE STRAIGHT LINE
				var lastPoint:Point = startPoint;
				//Draw any control points.				
				if (ctrls != null) {
					for (var i:uint=0; i<ctrls.length; i+=2){						
						if(!e.dashedLine){
							g.lineTo(ctrls[i]-x1, ctrls[i+1]-y1);
						}
						else{
							if(i ==0){
								drawDashedLine(startPoint.x, startPoint.y,ctrls[i]-x1, ctrls[i+1]-y1,e.dashLength,e.dashGapLength);
							}
							else{
								drawDashedLine(ctrls[i-2]-x1, ctrls[i-1]-y1,ctrls[i]-x1, ctrls[i+1]-y1,e.dashLength,e.dashGapLength);
							}
						}
						lastPoint.x = ctrls[i]-x1;
						lastPoint.y = ctrls[i+1]-y1;
					}
				}
				//Draw final line (or only one if no control points)
				if(!e.dashedLine){
					g.lineTo(endPoint.x,endPoint.y);
				}
				else{					
					drawDashedLine(lastPoint.x,lastPoint.y,endPoint.x,endPoint.y,e.dashLength,e.dashGapLength);
				}
			}
		}

		/**
		 * Draw a dashed line in graphics g
		 * Assumes relative placement on g, so starting from 0/0 		
		 *		
		 * @param startX	The init X
		 * @param startY	The init Y		
		 * @param endX		The final X
		 * @param endY		The final Y		
		 * @param dashLen	The length of the dash
		 * @param gapLen	The length of the gap
		 * 
		 */
		protected function drawDashedLine(startX:Number, startY:Number,endX:Number,endY:Number,dashLen:Number,gapLen:Number):void{
			var curX:Number = startX;
			var curY:Number = startY;			
			var lineOn:Boolean = true; //Is the line 'on', or off (dash or gap?)
			var xReached:Boolean = false;
			var yReached:Boolean = false;
			
			//trace('DRAW DASHED LINE TO ['+endX+','+endY+']');
			
			//Calc deltas
			var dx:Number = endX - startX;
			var dy:Number = endY - startY;
			var dd:Number = Math.sqrt(dx*dx + dy*dy);
			dx /= dd; 
			dy /= dd;
			
			//Start with filled line, keep drawing dash then gap until we run out of length in the line.
			while(Math.abs(curX) <= Math.abs(endX)  && (Math.abs(curY) <= Math.abs(endY) && !(curY==endY && curX==endX))){
				
				if(lineOn){
					//trace(' ['+endX+','+endY+'] lineOn CUR ['+curX+','+curY+']');					
					if(!xReached){
						curX += dx*dashLen;
						if(Math.abs(curX)>Math.abs(endX)){
							curX = endX;
						}
					}
					if(!yReached){
						curY += dy*dashLen;					
						if(Math.abs(curY)>Math.abs(endY)){
							curY = endY;
						}
					}					
					g.lineTo(curX,curY);		
					lineOn = false;
				}
				else{
					//trace(' ['+endX+','+endY+'] lineOff CUR ['+curX+','+curY+']');
					if(!xReached){
						curX += dx*gapLen;
						if(Math.abs(curX)>Math.abs(endX)){
							curX = endX;
						}
					}
					if(!yReached){
						curY += dy*gapLen;
						
						if(Math.abs(curY)>Math.abs(endY)){
							curY = endY;
						}
					}
					g.moveTo(curX,curY);
					lineOn = true;
				}
				
				if(curX == endX){xReached = true;}
				if(curY == endY){yReached = true;}
			}
			//trace(' ENDED AT: ['+curX+','+curY+']');
		}
		
	} // end of class EdgeRenderer
}