package flare.vis.operator.layout
{
	import flare.util.Orientation;
	import flare.vis.data.Data;
	import flare.vis.data.DataList;
	import flare.vis.data.EdgeMode;
	import flare.vis.data.NodeMode;
	
	/**
	 * Lays out nodes as by grouping defined in the dataset. 
	 * 
	 * A lot like Grouped layout, except there's clustering around the parents, 
	 * and also items not in the defined levels are then laid out clost to other things
	 * <br><br>
	 * Also, class diagrams work in reverse in terms of line direction. So children go up toward parents
	 * and this changes the code somewhat
	 * <br><br>
	 * And we have straight lines here rather than angled.
	 * 
	 * @author so33753
	 * 
	 */
	public class ClassDiagramLayout extends Layout
	{
		// -- Properties ------------------------------------------------------
		
		/** Property name for storing parameters for this layout. */
		public static const PARAMS:String = "classDiagramLayout";
		
		public static const LEFTTORIGHT:String = "leftToRight";
		public static const TOPTOBOTTOM:String = "topToBottom";
		
		private var _orient:String = Orientation.LEFT_TO_RIGHT; // orientation
		private var _bspace:Number = 20;  // the spacing between sibling nodes
		private var _tspace:Number = 25; // the spacing between subtrees
		private var _dspace:Number = 50; // the spacing between depth levels		
		private var _maxDepth:int = 0;
		
		/**
		 * An array of group names, defining the order in which to graph them.
		 */
		private var groupNames:Array;
		/**
		 * For each group, the last height or last width of placement for that group
		 * depending on if it's left to right or top to bottom layout 
		 */
		private var groupLastPos:Array;
		private var visitedNodes:Object;
		private var rootNodes:Array;		
		private var visData:Data;
		private var nodePositions:Object;
		
		//Levels and sizes per root node.
		private var levels:Object ;
		private var levelSizes:Object;
		
		//Edge Padding
		public var yOffset:Number = 50;
		public var xOffset:Number = 50;
			
		
		/** The orientation of the layout. */
		public function get orientation():String { return _orient; }
		public function set orientation(o:String):void { _orient = o; }
		
		/** The space between successive depth levels of the tree. */
		public function get depthSpacing():Number { return _dspace; }
		public function set depthSpacing(s:Number):void { _dspace = s; }
		
		/** The space between siblings in the tree. */
		public function get breadthSpacing():Number { return _bspace; }
		public function set breadthSpacing(s:Number):void { _bspace = s; }
		
		/** The space between different sub-trees. */
		public function get subtreeSpacing():Number { return _tspace; }
		public function set subtreeSpacing(s:Number):void { _tspace = s; }				
		
		// -- Methods ---------------------------------------------------------		
		/**
		 * Creates a new ClassDiagramLayout.
		 * @param orientation the orientation of the layout
		 * @param depthSpace the space between depth levels in the tree
		 * @param breadthSpace the space between siblings in the tree
		 * @param subtreeSpace the space between different sub-trees
		 * @param childWiggle the amount child nodes are shifted about their normal position to reduce line overlap 
		 * 					  between parent and child nodes (especially loop-back situation)
		 */		
		public function ClassDiagramLayout(	groupList:Array,orientation:String=Orientation.LEFT_TO_RIGHT, depthSpace:Number=100,
												breadthSpace:Number=20, subtreeSpace:Number=25){
			_orient = orientation;
			_dspace = depthSpace;
			_bspace = breadthSpace;
			_tspace = subtreeSpace;
			groupNames = groupList;			
		}
		
		private function orderDegree(a:NodeMode, b:NodeMode):int 
		{ 
			return a.degree<b.degree?1:a.degree>b.degree?-1:0;			
		} 
		
		/** @inheritDoc */
		protected override function layout():void
		{		
			width = xOffset;
			height = yOffset;			
			rootNodes = [];			
			visitedNodes = {};	
			groupLastPos = [];
			levels = {};
			levelSizes = {};
			nodePositions = {};
			
			var xPos:Number = xOffset;
			var yPos:Number = yOffset;
			var sizeTotal:Number = 0;	

			visData = visualization.data as Data;		
						
			if(visData.nodes.length == 0){
				return;
			}
			
			//Take first group as our starting point
			rootNodes = visData.group(groupNames[0]).toArray();

			//init group Last positions.
			var i:int;
			for(i=0;i<groupNames.length;i++){
				groupLastPos[i] = 0;
			}
								
			trace('============= CLASS LAYOUT START ============ HEIGHT IS: HEIGHT = '+height+' width='+width+' # rootNodes='+rootNodes.length);
			trace(groupNames);
			
			if(rootNodes.length > 0){
				//Position root nodes based on the size of the largest child group (may be any one of the levels)						
				//For correct spacing we can't just use the complete inNodes list though, as that can include nodes from many levels, we must do a pre-run through to 
				//get the sizes for each level, THEN use those to space things out.
				var nm:NodeMode;
				var innodes:Array;
				var cn:NodeMode;			
				var childStartX:Number;
				var childX:Number;
				var childStartY:Number;
				var childY:Number;
				var curGroup:DataList;										
				var curidx:int;
				
				for(i=0;i<rootNodes.length;i++){
					nm = rootNodes[i];
					levels[nm.data.uid] = [];
					levelSizes[nm.data.uid] = [];
					innodes = nm.inNodes;
					//WORK OUT LEVELS AND SIZES
					if(innodes.length > 0){
						//Lay out children as per groupings. 
						//Leave the levelSizes[uid][0] spot to be the maximum width of all children
						levelSizes[nm.data.uid][0] = 0;
						for(curidx=1;curidx<groupNames.length;curidx++){
							curGroup = visualization.data.group(groupNames[curidx]);
							levels[nm.data.uid][curidx] = [];
							levelSizes[nm.data.uid][curidx] = 0;
							for each(cn in innodes){
								if(curGroup.contains(cn)){
									levels[nm.data.uid][curidx].push(cn);
									if(orientation == LEFTTORIGHT){
										levelSizes[nm.data.uid][curidx] += cn.height + breadthSpacing;
									}
									else{								 
										levelSizes[nm.data.uid][curidx] += cn.width + breadthSpacing;
										//trace('LEVEL ['+curidx+'] Adding width ['+cn.width+'] total size ='+levelSizes[curidx]);
									}
								}							
							}
							
							//If this is the largest level we've seen, width wise
							if(levelSizes[nm.data.uid][curidx] > levelSizes[nm.data.uid][0]){
								levelSizes[nm.data.uid][0] = levelSizes[nm.data.uid][curidx] ;
							}
							//trace(nm.data.name+' level ['+curidx+'] size = '+levelSizes[nm.data.uid][curidx]+' MAX ='+levelSizes[nm.data.uid][0]);
						}
					}
				}
				
				//Place the root nodes
				var padding:Number = 0;
				for(i=0;i<rootNodes.length;i++){
					nm = rootNodes[i];
					
					if(orientation == LEFTTORIGHT){							
						yPos = height+(sizeTotal/2);
						if(i>0){
							yPos+=sizeTotal/(nm.inDegree*2) + subtreeSpacing;
						}
					}
					else{						
						//Position to the right of what's come before, making room 
						//for half the width of its largest sub level
						if(levelSizes[nm.data.uid][0]){
							padding = levelSizes[nm.data.uid][0]/2; 
						}
						xPos = width+padding;
						if(i>0){
							xPos+=subtreeSpacing;
						}
					}
					trace('POSITION ROOT NODE ['+i+'] '+nm.data.name+' at ['+xPos+','+yPos+'] through largest level being  ='+levelSizes[nm.data.uid][0]);
					placeNode(nm,xPos,yPos,0);	
					if(orientation == LEFTTORIGHT){		
						if(yPos + nm.height > groupLastPos[0]){
							groupLastPos[0] = yPos+nm.height;
						}
					}
					else{
						if(xPos + nm.width > groupLastPos[0]){
							groupLastPos[0] = xPos+nm.width;
						}
					}
				}
			}
			
			//We can have nodes that are left in the groups other than 'root' if they 
			//are not linked as such. Sooo, we have to lay them out in the correct places.			
			for(curidx=1;curidx<groupNames.length;curidx++){
				curGroup = visualization.data.group(groupNames[curidx]);
				var subGroupidx:int = 0;
				if(orientation == LEFTTORIGHT){
					xPos = width+depthSpacing;
				}
				else{
					yPos = height + depthSpacing;
				}
				for each(nm in curGroup){
					if(!visitedNodes[nm.data.uid]){
						subGroupidx++;
						if(orientation == LEFTTORIGHT){							
							yPos = groupLastPos[curidx]+breadthSpacing;
							groupLastPos[curidx] = yPos+nm.height;
							placeNode(nm,xPos,yPos,curidx);							
						}
						if(orientation == TOPTOBOTTOM){														
							xPos = groupLastPos[curidx]+breadthSpacing;
							groupLastPos[curidx] = xPos+nm.width;
							placeNode(nm,xPos,yPos,curidx);								
						}
					}
				}
			}
			
			styleEdges();
			
			updateEdgePoints(_t); 	// update edges
			super.layout(); 		//Set parent size			
			//trace('============= LAYOUT END ============ HEIGHT IS: HEIGHT = '+height+' width='+width);
		}					
		
		protected function placeNode(nm:NodeMode,nodeX:Number,nodeY:Number,nodeLevel:int):void
		{	
			//These layouts can include nodes more than once. If we've already been here once
			//then back away quietly.
			if(visitedNodes[nm.data.uid]){				
				return;
			}
			else{
				visitedNodes[nm.data.uid] = nm;
				nodePositions[nm.data.uid] = {x:nodeX,y:nodeY,level:nodeLevel}
			}			
			
			//Setting values by way of a transitioner, for animation purposes
			//trace('placeNode: '+nm.data.name+'['+nm.data.objectId+'] ('+nodeX+','+nodeY+')');
			var o:Object = _t.$(nm);
			//var eo:Object;
			var e:EdgeMode;
			var i:int;
			nm.h = _t.endSize(nm, _rect).height;
			
			o.x = nodeX;
			o.y = nodeY;
			
			// update layout size					
			if((o.x+nm.width) > width){
				width = o.x+nm.width;
			}
			if((o.y+nm.height) > height){
				height = o.y+nm.height;
			}
			
			//trace('--->'+n.data.name+'['+n.data.objectId+'] x='+o.x+' y='+o.y+' width='+width+' height='+height+' =-=-=-=-=-=-=-= My Children number '+n.outNodes.length)				
			
			//Traverse Groups, now seeing as with class diagrams the children link back to the parent, we need to work in reverse.
			//so, for each group, check all members and place those that have a link back to this node.
			
			//For correct spacing we can't just use the complete inNodes list though, as that can include nodes from many levels, we must do a pre-run through to 
			//get the sizes for each level, THEN use those to space things out.
			var innodes:Array = nm.inNodes;
			var cn:NodeMode;			
			var childStartX:Number;
			var childX:Number;
			var childStartY:Number;
			var childY:Number;
			var curGroup:DataList;
			var curidx:int;
						
			if(innodes.length > 0){								
				//Now we have just the nodes in levels for this parent, and
				//have the actual total sizes of each level, let's lay them oooout!
				var level:Array;
				var avgSize:Number;
				var levelSize:Number;
				var dif:Number
				
				//Only do levels below self
				for(curidx=nodeLevel+1;curidx<groupNames.length;curidx++){
					//If a pre-calculated level for root nodes, use that					
					if(levels[nm.data.uid]){
						level = levels[nm.data.uid][curidx];
						avgSize = level.length==0?0:levelSizes[nm.data.uid][curidx]/level.length;
						levelSize = level.length==0?0:levelSizes[nm.data.uid][curidx];
					}
					else{ //Else, just use entire level defs
						level = visData.group(groupNames[curidx]).toArray()
						avgSize = 0;
						for each (cn in level){
							if(orientation == LEFTTORIGHT){
								avgSize += cn.height;
							}
							else{
								avgSize += cn.width;
							}
						}
						levelSize = avgSize;
						avgSize = avgSize/level.length;
					}
					//trace('  Level ['+curidx+'] # Nodes='+level.length+' AVG='+avgSize);					
					for each(cn in level){
						//The check for an existence of an edge is only required for times we get in here for non root nodes with links											
						if(orientation == LEFTTORIGHT){
							childStartX = curidx*depthSpacing;
							childStartY = nodeY;			
							for(i=0;i<level.length;i++){						
								cn =innodes[i];												
								childY = (childStartY - ((level.length-1)*(avgSize/2) )+(i*breadthSpacing));																	
								
								placeNode(cn,childStartX,childY,curidx);
								
								//Set Control point to make straight rather than angled lines:
							/*	e = visData.findEdgeFor(cn,nm);																		
								dif = -(childY-nodeY);
								//trace('Calc control points for '+n.data.name+'['+nodeX+','+nodeY+'] '+cn.data.name+'['+childX+','+childY+'] DIF='+dif);
								
								//NEEDS DOING!
								e.points = [childX,childY-((childY-nodeY)/2),childX-dif,childY-((childY-nodeY)/2)];
								*/
								if(childY+cn.height > groupLastPos[curidx]){
									groupLastPos[curidx]  = childY+cn.height;
								}
							}									
						}
						if(orientation == TOPTOBOTTOM){
							childStartY = curidx*depthSpacing;
							childX = nodeX + (avgSize/2) - (levelSize/2);			
							
							for(i=0;i<level.length;i++){						
								cn = level[i];											
								if(i>0){
									childX += breadthSpacing+avgSize;
								}
								childY = childStartY;
								//trace('   ChildNode : '+cn.data.name+' : Group ['+groupNames[curidx]+'] X='+childX+' Y='+childY);									
								placeNode(cn,childX,childY,curidx);

								if(childX+cn.width > groupLastPos[curidx]){
									groupLastPos[curidx]  = childX+cn.width;
								}
							}					
						}						
					}
				}									
			}									
		}
		

		/**
		 * Edge styling is thus:<br> 
		 *  If edge from child to parent level, then add control
		 * 
		 */
		private function styleEdges():void{
			var s:NodeMode;
			var t:NodeMode;
			var sPos:Object;
			var tPos:Object;
			var dif:Number;
			var yMidPoint:Number;
			
			for each(var e:EdgeMode in visData.edges){	
				s = e.source;
				sPos = nodePositions[s.data.uid];
				t = e.target;
				tPos = nodePositions[t.data.uid];
				
				//trace('EDGE FROM '+s.data.name+' L: '+sPos.level+' ['+sPos.x+','+sPos.y+'] TO '+t.data.name+' L:'+tPos.level+' ['+tPos.x+','+tPos.y+']');
				//only use control points for nodes on different levels.									
				if(sPos.level > tPos.level){					
					dif = -(sPos.x-tPos.x);//childX<nodeX?-(childX-nodeX):-(nodeX-childX);
					//Set offsets to keep away from parent node, and to shift each incoming line apart from each other.
					e.endYOffset = 30;
					//To avoid all lines from parent to child running over each other, offset the end XPos of each incoming
					//line by it's relative position along the x position as a fraction of the width of the target node.
					e.endXOffset = dif/levelSizes[t.data.uid][sPos.level]*t.width; 	
					
					//We have issues of the horizontal portions of 
					//the child lines running over each other, so this needs to stagger where
					//they lie
					yMidPoint = sPos.y-((sPos.y-tPos.y)/2) + e.endXOffset;				
					e.points = [sPos.x,yMidPoint,sPos.x-dif+e.endXOffset,yMidPoint];
				}
				else{
					e.dashedLine = true;
					e.dashLength = 2;
					e.dashGapLength = 4;
				}
			}
		}				
		
		private function setVisibility(n:NodeMode, o:Object, visible:Boolean):void
		{
			o.alpha = visible ? 1.0 : 0.0;
			o.mouseEnabled = visible;
			if (n.parentEdge != null) {
				o = _t.$(n.parentEdge);
				o.alpha = visible ? 1.0 : 0.0;
				o.mouseEnabled = visible;
			}			
		}
		
		private function setSizes(n:NodeMode):void
		{
			_t.endSize(n, _rect);
			n.w = _rect.width;
			n.h = _rect.height;
		}
		
		private function spacing(l:NodeMode, r:NodeMode, siblings:Boolean):Number
		{
			var w:Boolean = Orientation.isVertical(_orient);
			return (siblings ? _bspace : _tspace) + 0.5 *
				(w ? l.w + r.w : l.h + r.h)
		}		
		// -- Parameter Access ------------------------------------------------
		
		private function params(n:NodeMode):Params
		{
			var p:Params = n.props[PARAMS] as Params;
			if (p == null) {
				p = new Params();
				n.props[PARAMS] = p;
			}
			if (p.number == -2) { p.init(n); }
			return p;
		}		
	} // end of class NodeLinkTreeLayout	
}


import flare.vis.data.NodeMode;

class Params {
	public var prelim:Number = 0;
	public var mod:Number = 0;
	public var shift:Number = 0;
	public var change:Number = 0;
	public var number:int = -2;
	public var ancestor:NodeMode = null;
	public var thread:NodeMode = null;
	
	public function init(item:NodeMode):void
	{
		ancestor = item;
		number = -1;
	}
	
	public function clear():void
	{
		number = -2;
		prelim = mod = shift = change = 0;
		ancestor = thread = null;
	}
} // end of class Params