package flare.vis.operator.layout
{
	import flare.util.Orientation;
	import flare.vis.data.Data;
	import flare.vis.data.DataList;
	import flare.vis.data.EdgeMode;
	import flare.vis.data.NodeMode;

	/**
	 * Lays out nodes as by grouping defined in the dataset. 
	 * Define groups  
	 * @author so33753
	 * 
	 */
	public class GroupedNodeLinkLayout extends Layout
	{
		// -- Properties ------------------------------------------------------
		
		/** Property name for storing parameters for this layout. */
		public static const PARAMS:String = "groupedNodeLinkLayout";
		
		public static const LEFTTORIGHT:String = "leftToRight";
		public static const TOPTOBOTTOM:String = "topToBottom";
		
		private var _orient:String = Orientation.LEFT_TO_RIGHT; // orientation
		private var _bspace:Number = 5;  // the spacing between sibling nodes
		private var _tspace:Number = 25; // the spacing between subtrees
		private var _dspace:Number = 50; // the spacing between depth levels		
		private var _maxDepth:int = 0;

		/**
		 * An array of group names, defining the order in which to graph them.
		 */
		private var groupNames:Array;
		/**
		 * For each group, the last height or last width of placement for that group
		 * depending on if it's left to right or top to bottom layout 
		 */
		private var groupLastPos:Array;
		private var visitedNodes:Object;
		private var rootNodes:Array;
		
		
		
		//Padding
		public var yOffset:Number = 50;
		public var xOffset:Number = 50;
		
		/** The orientation of the layout. */
		public function get orientation():String { return _orient; }
		public function set orientation(o:String):void { _orient = o; }
		
		/** The space between successive depth levels of the tree. */
		public function get depthSpacing():Number { return _dspace; }
		public function set depthSpacing(s:Number):void { _dspace = s; }
		
		/** The space between siblings in the tree. */
		public function get breadthSpacing():Number { return _bspace; }
		public function set breadthSpacing(s:Number):void { _bspace = s; }
		
		/** The space between different sub-trees. */
		public function get subtreeSpacing():Number { return _tspace; }
		public function set subtreeSpacing(s:Number):void { _tspace = s; }
		
		public var childWiggle:Number;
				
		// -- Methods ---------------------------------------------------------		
		/**
		 * Creates a new GroupedNodeLinkLayout.
		 * @param orientation the orientation of the layout
		 * @param depthSpace the space between depth levels in the tree
		 * @param breadthSpace the space between siblings in the tree
		 * @param subtreeSpace the space between different sub-trees
		 * @param childWiggle the amount child nodes are shifted about their normal position to reduce line overlap 
		 * 					  between parent and child nodes (especially loop-back situation)
		 */		
		public function GroupedNodeLinkLayout(	groupList:Array,orientation:String=Orientation.LEFT_TO_RIGHT, depthSpace:Number=270,
												breadthSpace:Number=50, subtreeSpace:Number=25,childWiggleVal:Number=25){
			_orient = orientation;
			_dspace = depthSpace;
			_bspace = breadthSpace;
			_tspace = subtreeSpace;
			groupNames = groupList;
			childWiggle = childWiggleVal;
		}
		
		private function orderDegree(a:NodeMode, b:NodeMode):int 
		{ 
			return a.degree<b.degree?1:a.degree>b.degree?-1:0;			
		} 
		
		/** @inheritDoc */
		protected override function layout():void
		{
			
			width = xOffset;
			height = yOffset;			
			rootNodes = [];			
			visitedNodes = {};	
			groupLastPos = [];
			var nm:NodeMode;
			var data:Data = visualization.data;
			
			trace('============= GROUPED LAYOUT START ============ HEIGHT IS: HEIGHT = '+height+' width='+width);
			
			//Take first group as our starting point
					
			if(data.nodes.length == 0){
				return;
			}
			rootNodes = data.group(groupNames[0]).toArray();

			//Quit if we have no root nodes
			if(rootNodes.length == 0){
				return;
			}
			
			//initiate the highest degree node (most incoming/outgoing links)
			rootNodes.sort(orderDegree);											
			
			var xPos:Number;
			var yPos:Number
			
			//Init to start points that won't have nodes off screen.
			nm = rootNodes[0];
			
			if(orientation == LEFTTORIGHT){
				yPos = yOffset+((nm.outDegree-1)*(nm.height/2));
				xPos = xOffset;
			}
			else{
				xPos = xOffset+((nm.outDegree-1)*(nm.width/2));
				yPos = yOffset;
			}
			
			//init group Last positions.
			var i:int;
			for(i=0;i<groupNames.length;i++){
				groupLastPos[i] = 0;
			}
			
			//Position root nodes
			for(i=0;i<rootNodes.length;i++){
				nm = rootNodes[i];
					
				if(orientation == LEFTTORIGHT){					
					yPos = height+((nm.outDegree)*(nm.height/2));
					if(i>0){
						yPos+=subtreeSpacing;
					}
				}
				else{				
					xPos = width+((nm.outDegree)*(nm.width/2));
					if(i>0){
						xPos+=subtreeSpacing;
					}
				}
				placeNode(nm,xPos,yPos);	
				if(orientation == LEFTTORIGHT){		
					if(yPos + nm.height > groupLastPos[0]){
						groupLastPos[0] = yPos+nm.height;
					}
				}
				else{
					if(xPos + nm.width > groupLastPos[0]){
						groupLastPos[0] = xPos+nm.width;
					}
				}
			}	
			
			
			
			//We can have nodes that are left in the groups other than 'root' if they 
			//are not linked as such. Sooo, we have to lay them out in the correct places.			
			for(var curidx:int=0;curidx<groupNames.length;curidx++){
				var curGroup:DataList = visualization.data.group(groupNames[curidx]);
				var subGroupidx:int = 0;
				for each(nm in curGroup){
					if(!visitedNodes[nm.data.uid]){
						subGroupidx++;
						if(orientation == LEFTTORIGHT){
							xPos = curidx*depthSpacing;	
							yPos = groupLastPos[curidx]+breadthSpacing;
							groupLastPos[curidx] = yPos+nm.height;
							placeNode(nm,xPos,yPos);
														
						}
						if(orientation == TOPTOBOTTOM){							
							yPos = curidx*depthSpacing;	
							xPos = groupLastPos[curidx]+breadthSpacing;
							groupLastPos[curidx] = xPos+nm.width;
							placeNode(nm,xPos,yPos);								
						}
					}
				}
			}
					
			updateEdgePoints(_t); 	// update edges
			super.layout(); 		//Set parent size			
			//trace('============= LAYOUT END ============ HEIGHT IS: HEIGHT = '+height+' width='+width);
		}					
					
		protected function placeNode(n:NodeMode,nodeX:Number,nodeY:Number):void
		{	
			//These layouts can include nodes more than once. If we've already been here once
			//then back away quietly.
			if(visitedNodes[n.data.uid]){
				return;
			}
			else{
				visitedNodes[n.data.uid] = n;
			}
			
			
			//Setting values by way of a transitioner, for animation purposes
			//trace('placeNode: '+n.data.name+'['+n.data.objectId+']');
			var o:Object = _t.$(n);
			var i:int;
			n.h = _t.endSize(n, _rect).height;
			
			o.x = nodeX;
			o.y = nodeY;
								
			// update layout size					
			if((o.x+n.width) > width){
				width = o.x+n.width;
			}
			if((o.y+n.height) > height){
				height = o.y+n.height;
			}

			//trace('--->'+n.data.name+'['+n.data.objectId+'] x='+o.x+' y='+o.y+' width='+width+' height='+height+' =-=-=-=-=-=-=-= My Children number '+n.outNodes.length)				
			
			//Traverse children 
			var on:Array = n.outNodes;
			var cn:NodeMode;			
			var childX:Number;
			var childY:Number;
			var curGroup:DataList;
			var curidx:int;
			if(on.length > 0){
				//Lay out children as per groupings. Always go through all groups as objects can link
				//back to higher levels. Left to Right will have groups in vertical lines, Top to Bottom
				//will have them in horizontal lines
				for(curidx=0;curidx<groupNames.length;curidx++){
					curGroup = visualization.data.group(groupNames[curidx]);
					for each(cn in on){
						if(curGroup.contains(cn)){
							if(orientation == LEFTTORIGHT){
								childX = curidx*depthSpacing;
								childY = o.y;			
								for(i=0;i<on.length;i++){						
									cn =on[i];												
									//trace(' X='+childX+' Y='+((childY - ((on.length-1)*(n.height/2) )+(i*breadthSpacing))+childWiggleAmt));
									placeNode(cn,childX,(childY - ((on.length-1)*(n.height/2) )+(i*breadthSpacing)));
									if((childY - ((on.length-1)*(n.height/2) )+(i*breadthSpacing))+cn.height > groupLastPos[curidx]){
										groupLastPos[curidx]  = (childY - ((on.length-1)*(n.height/2) )+(i*breadthSpacing))+cn.height;
									}
								}									
							}
							if(orientation == TOPTOBOTTOM){
								childY = curidx*depthSpacing;
								childX = o.x;			
								for(i=0;i<on.length;i++){						
									cn =on[i];												
									//trace(' X='+childX+' Y='+((childY - ((on.length-1)*(n.height/2) )+(i*breadthSpacing))+childWiggleAmt));
									placeNode(cn,(childX - ((on.length-1)*(n.width/2) )+(i*breadthSpacing)),childY);
									if((childX - ((on.length-1)*(n.width/2) )+(i*breadthSpacing))+cn.width > groupLastPos[curidx]){
										groupLastPos[curidx]  = (childX - ((on.length-1)*(n.width/2) )+(i*breadthSpacing))+cn.width;
									}
								}					
							}
						}
					}
					
				}
			}
		}
		
		private function setVisibility(n:NodeMode, o:Object, visible:Boolean):void
		{
			o.alpha = visible ? 1.0 : 0.0;
			o.mouseEnabled = visible;
			if (n.parentEdge != null) {
				o = _t.$(n.parentEdge);
				o.alpha = visible ? 1.0 : 0.0;
				o.mouseEnabled = visible;
			}			
		}
		
		private function setSizes(n:NodeMode):void
		{
			_t.endSize(n, _rect);
			n.w = _rect.width;
			n.h = _rect.height;
		}
		
		private function spacing(l:NodeMode, r:NodeMode, siblings:Boolean):Number
		{
			var w:Boolean = Orientation.isVertical(_orient);
			return (siblings ? _bspace : _tspace) + 0.5 *
				(w ? l.w + r.w : l.h + r.h)
		}		
		// -- Parameter Access ------------------------------------------------
		
		private function params(n:NodeMode):Params
		{
			var p:Params = n.props[PARAMS] as Params;
			if (p == null) {
				p = new Params();
				n.props[PARAMS] = p;
			}
			if (p.number == -2) { p.init(n); }
			return p;
		}		
	} // end of class NodeLinkTreeLayout	
}


import flare.vis.data.NodeMode;

class Params {
	public var prelim:Number = 0;
	public var mod:Number = 0;
	public var shift:Number = 0;
	public var change:Number = 0;
	public var number:int = -2;
	public var ancestor:NodeMode = null;
	public var thread:NodeMode = null;
	
	public function init(item:NodeMode):void
	{
		ancestor = item;
		number = -1;
	}
	
	public function clear():void
	{
		number = -2;
		prelim = mod = shift = change = 0;
		ancestor = thread = null;
	}
} // end of class Params