package flare.vis.operator.layout
{
	import flare.util.Arrays;
	import flare.vis.data.Data;
	import flare.vis.data.NodeMode;
	
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	/**
	 * Layout that places tree nodes in a radial layout, laying out depths of a tree
	 * along circles of increasing radius. 
	 * This layout can be used for both node-link diagrams, where nodes are
	 * connected by edges, and for radial space-filling ("sunburst") diagrams.
	 * To generate space-filling layouts, nodes should have their shape
	 * property set to <code>Shapes.WEDGE</code> and the layout instance should
	 * have the <code>useNodeSize<code> property set to false.
	 * 
	 * <p>The algorithm used is an adaptation of a technique by Ka-Ping Yee,
	 * Danyel Fisher, Rachna Dhamija, and Marti Hearst, published in the paper
	 * <a href="http://citeseer.ist.psu.edu/448292.html">Animated Exploration of
	 * Dynamic Graphs with Radial Layout</a>, InfoVis 2001. This algorithm computes
	 * a radial layout which factors in possible variation in sizes, and maintains
	 * both orientation and ordering constraints to facilitate smooth and
	 * understandable transitions between layout configurations.
	 * </p>
	 * 
	 * <p>
	 * This Snowflake layout is designed for use in database schema display and takes
	 * the outDegree/outNodes and replaces them in inDegree/inNodes as children 
	 * </p>
	 */
	public class SnowflakeLayout extends Layout
	{
		// -- Properties ------------------------------------------------------
		
		/** Property name for storing parameters for this layout. */
		public static const PARAMS:String = "snowflakeLayoutLayoutParams";
		/** The default radius increment between depth levels. */
		public static const DEFAULT_RADIUS:int = 10;
	
	    private var _maxDepth:int = 0;
	    private var _radiusInc:Number = DEFAULT_RADIUS;
	    private var _theta1:Number = Math.PI/2;
	    private var _theta2:Number = Math.PI/2 - 2*Math.PI;
	    private var _sortAngles:Boolean = true;
	    private var _setTheta:Boolean = false;
	    private var _autoScale:Boolean = false;
	    private var _useNodeSize:Boolean = false;
	    private var _prevRoot:NodeMode = null;
		private var _starSpacing:Number = 100;
		
		//New vars
		private var visitedNodes:Object;
		

		/** The radius increment between depth levels. */
		public function get radiusIncrement():Number { return _radiusInc; }
		public function set radiusIncrement(r:Number):void { _radiusInc = r; }
		
		/** Flag determining if nodes should be sorted by angles to help
		 *  maintain ordering across different spanning-tree configurations.
		 *  This sorting is important for understandable transitions when using
		 *  a radial layout of a general graph. However, it is unnecessary for
		 *  tree structures and increases the running time of layout. */
		public function get sortAngles():Boolean { return _sortAngles; }
		public function set sortAngles(b:Boolean):void { _sortAngles = b; }
		
		/** Flag indicating if the layout should automatically be scaled to
		 *  fit within the layout bounds. */
		public function get autoScale():Boolean { return _autoScale; }
		public function set autoScale(b:Boolean):void { _autoScale = b; }
		
		/** The initial angle for the radial layout (in radians). */
		public function get startAngle():Number { return _theta1; }
		public function set startAngle(a:Number):void {
			_theta2 += (a - _theta1);
			_theta1 = a;
			_setTheta = true;
		}
		
		/** The angular width of the layout (in radians, default is 2 pi). */
		public function get angleWidth():Number { return _theta1 - _theta2; }
		public function set angleWidth(w:Number):void {
			_theta2 = _theta1 - w;
			_setTheta = true;
		}

		/** Flag indicating if node's <code>size</code> property should be
		 *  used to determine layout spacing. If a space-filling radial
		 *  layout is desired, this value must be false for the layout
		 *  to be accurate. */
		public function get useNodeSize():Boolean { return _useNodeSize; }
		public function set useNodeSize(b:Boolean):void {
			_useNodeSize = b;
		}

		/**
		 * How much space to put between stars. 
		 * @return 
		 * 
		 */
		public function get starSpacing():Number { return _starSpacing; }
		public function set starSpacing(s:Number):void {
			_starSpacing = s;
		}
		
		/**
		 * Extra padding around calculated node sizes to add to ensure nodes
		 * don't overlap in their placement. 
		 */		
		public var nodePadding:Number = 0;
		
		// -- Methods ---------------------------------------------------------

		/**
		 * Creates a new SnowflakeLayout.
		 * @param radius the radius increment between depth levels
		 * @param sortAngles flag indicating if nodes should be sorted by angle
		 *  to maintain node ordering across spanning-tree configurations
		 * @param autoScale flag indicating if the layout should automatically
		 *  be scaled to fit within the layout bounds
		 */		
		public function SnowflakeLayout(radius:Number=DEFAULT_RADIUS,starSpacing:Number=100,
										sortAngles:Boolean=true, autoScale:Boolean=false)
		{			
			layoutType = POLAR;
			_radiusInc = radius;
			_sortAngles = sortAngles;
			_autoScale = autoScale;
			_starSpacing = starSpacing;
		}

		/** @inheritDoc */
		protected override function layout():void
		{
			var rootNodes:Array = [];
			var visData:Data = visualization.data as Data;
			var n:NodeMode;					 			
			var totalWidth:Number = 0;
			
			//Init vars
			_anchor = new Point(0,0);			
			width = 0;//visualization.width;
			height = 0;//visualization.height;
			visitedNodes = {};			
			
			trace('SNOWFLAKE LAYOUT START ==============');
			//First, look for a datagroup within the dataset called rootNodes. These are nodes that have been tagged as being start 
			//points for graphing.
			if(visData.group('rootNodes')){						
				rootNodes = visData.group('rootNodes').toArray();				
			}
			//Quick fix, if no root nodes, just make all of them root...
			if(rootNodes.length == 0){
				rootNodes = visData.nodes.toArray();
			}									
						
			//var n:NodeMode = layoutRoot as NodeMode;
			//trace('SNOWFLAKE layout, # root nodes = '+rootNodes.length+' RADIUS INC='+radiusIncrement);
			for each(n in rootNodes){ 
				totalWidth = placeNode(n,totalWidth);	
			}
			
			trace('-=-= ORPHAN NODES =-=-');		
			//Now, we may have items to draw that are not attached to a star
			//We'll place them along the bottom
			var orphanAnchor:Point = new Point(0,height+50);
			for each(n in visData.nodes){
				if(!visitedNodes[n.data.uid]){
					orphanAnchor.x+=n.width/2;
					totalWidth = placeNode(n,totalWidth,orphanAnchor);
					orphanAnchor.x +=  50;
				}
			}

			updateEdgePoints(_t);
			
			trace('SNOWFLAKE LAYOUT END ============== HEIGHT/WIDTH = '+height+'/'+width);
			super.layout(); 	
		}
		
		private function placeNode(n:NodeMode,totalWidth:Number,overrideAnchor:Point=null):Number{
			var angWidth:Number = 0;
		
			if (n == null) { 
				_t = null; 
				return totalWidth; 
			}
			var np:Params = params(n);
			
			// calc relative widths and maximum tree depth
			// performs one pass over the tree
			_maxDepth = 0;
			angWidth = calcAngularWidth(n, 0); 
							
			totalWidth += angWidth +_maxDepth*radiusIncrement;
			
			if (_autoScale) setScale(layoutBounds);
			if (!_setTheta) calcAngularBounds(n);
			
			//Put each star's root at the middle vertically,
			//and half their own width in from the right edge each time
			//this should make them layout left to right.
			if(!overrideAnchor){
				trace('ROOT '+n.data.name+' ANGULAR WIDTH='+angWidth+' DEPTH='+_maxDepth);
				_anchor.y = layoutBounds.height/2;			
				_anchor.x = totalWidth - angWidth + _starSpacing;
			}
			else{
				_anchor = overrideAnchor;	
			}
			
			trace('ROOTNDE '+n.data.name+' ANGULAR WIDTH= '+ angWidth+' TOTALWIDTH='+totalWidth+' PLACING AT X,Y='+_anchor.x+','+_anchor.y); 
			
			// perform the layout			
			if (_maxDepth > 0) {
				trace('LAYOUT CALLING doLayout with r='+_radiusInc);
				doLayout(n, _radiusInc, _theta1, _theta2);
			} else if (n.inDegree > 0) {
				n.visitNodes(function(n:NodeMode):void {
					n.origin = _anchor;
					var o:Object = _t.$(n);
					// collapse to inner radius						
					o.radius = o.h = o.v = _radiusInc / 2;
					trace('VISIT NODE '+n.data.name+' radius = '+o.radius);
					o.alpha = 0;
					o.mouseEnabled = false;
					
					/*if (n.firstInEdge != null)
					_t.$(n.firstInEdge).alpha = false;*/
				},NodeMode.IN_LINKS);
			}
			
			// update properties of the root node
			np.angle = _theta2 - _theta1;
			n.origin = _anchor;
			update(n, 0, _theta1+np.angle/2, np.angle, true);
			if (!_t.immediate) {
				delete _t._(n).values.radius;
				delete _t._(n).values.angle;
			}
			_t.$(n).x = _anchor.x;
			_t.$(n).y = _anchor.y;			
			//trace('ANCHOR x='+ _anchor.x +' y='+ _anchor.y+' TOTALWIDTH ='+totalWidth);		
			return totalWidth;
		}				

		private function setScale(bounds:Rectangle):void
		{
			trace('SET SCALE!!!');
	        var r:Number = Math.min(bounds.width, bounds.height)/2.0;
	        if (_maxDepth > 0) _radiusInc = r / _maxDepth;
	    }
		
	    /**
	     * Calculates the angular bounds of the layout, attempting to
	     * preserve the angular orientation of the display across transitions.
	     */
	    private function calcAngularBounds(r:NodeMode):void
	    {
			//trace('calcAngularBounds '+r.data.name);
	        if (_prevRoot == null || r == _prevRoot)
	        {
				//trace('No prev root, return');
	            _prevRoot = r; return;
	        }
	        
	        // try to find previous parent of root
	        var p:NodeMode = _prevRoot;
			var pp:NodeMode;
			
	        while (true) {
	        	pp = p.firstOutNode;
	            if (pp == r) {
	                break;
	            } else if (pp == null) {
	                _prevRoot = r;
	                return;
	            }
	            p = pp;
	        }
	
	        // compute offset due to children's angular width
	        var dt:Number = 0;
	        
	        for each (var n:NodeMode in sortedChildren(r)) {
	        	if (n == p) break;
	        	dt += params(n).width;
	        }
	        
	        var rw:Number = params(r).width;
	        var pw:Number = params(p).width;
	        dt = -2*Math.PI * (dt+pw/2)/rw;
	
	        // set angular bounds
	        _theta1 = dt + Math.atan2(p.y-r.y, p.x-r.x);
	        _theta2 = _theta1 + 2*Math.PI;
	        _prevRoot = r;     
	    }
		
		/**
	     * Computes relative measures of the angular widths of each
	     * expanded subtree. Node diameters are taken into account
	     * to improve space allocation for variable-sized nodes.
	     * 
	     * This method also updates the base angle value for nodes 
	     * to ensure proper ordering of nodes.
	     */
	    private function calcAngularWidth(n:NodeMode, d:int):Number
	    {
			trace('	calcAngularWidth ['+n.data.name+'] Depth='+d)			
			if (d > _maxDepth) _maxDepth = d;       
	        var aw:Number = 0, diameter:Number = 0;
	        if (_useNodeSize && d > 0) {
	        	//diameter = 1;
	        	diameter = n.expanded && n.inDegree > 0 ? 0 : _t.$(n).size;
	        } else if (d > 0) {
	        	var w:Number = n.width;
				var h:Number = n.height;
	        	diameter = Math.sqrt(w*w+h*h)/d;					
	        	if (isNaN(diameter)) diameter = 0;				
	        }

			//trace('calcAngularWidth ['+n.data.name+'] d='+d+' maxDepth='+_maxDepth+' EXP? '+n.expanded+' outDegree='+n.outDegree+' inDegree='+n.inDegree);
	        if (n.expanded && n.inDegree > 0) {
				var c:NodeMode
	        	for (var i:int=0;i<n.inNodes.length;i++){
					c = n.inNodes[i];				
					aw += calcAngularWidth(c, d+1);
					trace('   Adding child width, now aw='+aw+'  '+c.data.name);
	        	}
	        	aw = Math.max(diameter, aw);
	        } else {
	        	aw = diameter;
	        }
			//Add the ol' Node Padding
			aw+=nodePadding;
			params(n).width = aw;
			//trace(n.data.name+'  DIAMETER ='+diameter+'  ANGLULAR WIDTH='+aw);
	        return aw;
	    }
		
		private static function normalize(angle:Number):Number
		{
	        while (angle > 2*Math.PI)
	            angle -= 2*Math.PI;
	        while (angle < 0)
	            angle += 2*Math.PI;
	        return angle;
	    }

		private function sortedChildren(n:NodeMode):Array
		{
			var cc:int = n.inDegree;
			if (cc == 0) return Arrays.EMPTY;
			var angles:Array = new Array(cc);
	        
	        if (_sortAngles) {
	        	// update base angle for node ordering			
				var base:Number = -_theta1;
				var p:NodeMode = n.firstInNode;
	        	if (p != null) base = normalize(Math.atan2(p.y-n.y, n.x-p.x));
	        	
	        	// collect the angles
	        	var c:NodeMode;
		        for (var i:uint=0; i<n.inNodes.length;i++) {
					c = n.inNodes[i];
		        	angles[i] = normalize(-base + Math.atan2(c.y-n.y,n.x-c.x));
		        }
		        // get array of indices, sorted by angle
		        angles = angles.sort(Array.NUMERIC | Array.RETURNINDEXEDARRAY);
		        // switch in the actual nodes and return
		        for (i=0; i<cc; ++i) {
		        	angles[i] = n.getInNode(angles[i]);
		        }
		    } else {
		    	for (i=0; i<cc; ++i) {
		        	angles[i] = n.getInNode(i);
		        }
		    }
	        
	        return angles;
	    }
		
		/**
	     * Compute the layout.
	     * @param n the root of the current subtree under consideration
	     * @param r the radius, current distance from the center
	     * @param theta1 the start (in radians) of this subtree's angular region
	     * @param theta2 the end (in radians) of this subtree's angular region
	     */
	    private function doLayout(n:NodeMode, r:Number,theta1:Number, theta2:Number):void
	    {			
			
	    	var dtheta:Number = theta2 - theta1;
	    	var dtheta2:Number = dtheta / 2.0;
	    	var width:Number = params(n).width; //this param would be Angular Width
	    	var cfrac:Number;
			var nfrac:Number = 0;
	      	
			//trace('doLayout '+n.data.name+' R='+r+' theta1='+theta1+' theta2='+theta2+' dtheta='+dtheta);
			
			// trace('== doLayout : '+n.data.name) 
	        for each (var c:NodeMode in sortedChildren(n)) {	        	
				var cp:Params = params(c);
	            cfrac = cp.width / width; //The fraction of the parent angular width that the child represents.
	            if (c.expanded && c.inDegree > 0)
	            {
					//trace('  '+n.data.name+' --- dl subcall to >'+c.data.name)
	                doLayout(c, r+_radiusInc, theta1 + nfrac*dtheta, 
	                                          theta1 + (nfrac+cfrac)*dtheta);
	            }
	            else if (c.inDegree > 0)
	            {
	            	var cr:Number = r + _radiusInc;
	            	var ca:Number = theta1 + nfrac*dtheta + cfrac*dtheta2;
					//trace('  --- visitNodes >'+c.data.name)
	            	c.visitNodes(function(n:NodeMode):void {
						//trace('DL visitNode >'+n.data.name);
	            		n.origin = _anchor;
	            		update(n, cr, minAngle(n.angle, ca), 0, false);
	            	},NodeMode.IN_LINKS);
	            }
	            
	            c.origin = _anchor;
	            var a:Number = minAngle(c.angle, theta1 + nfrac*dtheta + cfrac*dtheta2);
	            cp.angle = cfrac * dtheta;
				//trace('  -- '+n.data.name+':  UPDATING CHILD FRACTION = '+cfrac+' USING r='+r+' a='+a+' cp.angle='+cp.angle+'  COMPARE '+c.angle+'->'+(theta1 + nfrac*dtheta + cfrac*dtheta2)+' | '+c.data.name);
				
	            update(c, r, a, cp.angle, true);
	            nfrac += cfrac;
	        }
	    }
		
		/**
		 * 
		 * @param n The Node
		 * @param r Radius
		 * @param a Angle
		 * @param aw Angular width
		 * @param v Is visible?
		 * 
		 */
		private function update(n:NodeMode, r:Number, a:Number, aw:Number, v:Boolean) : void{
			if(!visitedNodes[n.data.uid]){
				visitedNodes[n.data.uid] = n;
			}
		
			var o:Object = _t.$(n);
			var alpha:Number = v ? 1 : 0;
			o.radius = r;
			o.angle = a;					
			
			//trace('    -- Node update  r='+r+' a='+a+' aw='+aw+'  New X='+ (r * Math.cos(a) + n.origin.x)+' Y='+(-r * Math.sin(a) + n.origin.y)+' ['+n.data.name+']');			
			//Set layout height and width if this is nudging an edge
			if(n.width + (r * Math.cos(a) + n.origin.x) > width){
				width = n.width + (r * Math.cos(a) + n.origin.x);
			}
			if(n.height + (-r * Math.sin(a) + n.origin.y) > height){
				height = n.height + (-r * Math.sin(a) + n.origin.y);
			}
			
			if (aw == 0) {
				o.h = o.v = r - _radiusInc/2;
			} else {
				o.h = r + _radiusInc/2;
				o.v = r - _radiusInc/2;
			}			
			o.w = aw;
			o.u = a - aw/2;
			o.alpha = alpha;
			o.mouseEnabled = v;
		/*	What IS this doing?
			if (n.firstInEdge != null)
				_t.$(n.firstInEdge).alpha = alpha;
			*/
		}
				
		private function params(n:NodeMode):Params
		{
			var p:Params = n.props[PARAMS];
			if (p == null) {
				p = new Params();
				n.props[PARAMS] = p;
			}
			return p;
		}
		
	}
}

class Params {
	public var width:Number = 0;
	public var angle:Number = 0;
}